/*
 * $Id: client.c,v 1.2 2003/02/21 12:58:59 tsavola Exp $
 * Timo Savola <timo.savola@movial.fi>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "client.h"
#include "protocol.h"
#include "auth.h"

#ifdef NDEBUG
#define DEBUG(msg)
#else
#define DEBUG(msg) printf("%s\n", msg)
#endif

/* The environment of the process. */
extern char **environ;

/** The socket. */
static int sd;

/** Buffer used in send_packet() and receive_packet(). */
static char buf[BUFSIZE];

/** The return code that will eventually be received from the server. */
static int16_t rc = 0;

/**
 * Writes a STDIN or STDIN_EOF packet to the socket (sd).
 * @return 0 on success, -1 on error
 */
static int send_packet(void)
{
	int count;

	count = read(STDIN_FILENO, &buf, BUFSIZE);
	if (count < 0) {
		perror("Can't read from stdin");
		return -1;
	}

	if (count > 0) {
		if (write_packet(sd, PTYPE_STDIN, &buf, count) < 0) {
			perror("Can't write STDIN packet to socket");
			return -1;
		}
	} else {
		if (write_packet(sd, PTYPE_STDIN_EOF, NULL, 0) < 0) {
			perror("Can't write STDIN_EOF packet to socket");
			return -1;
		}
	}

	return 0;
}

/**
 * Reads the contents of a STDOUT or STDERR packet from the socket (sd) into
 * the buffer (buf).
 * @param size to be read
 * @param fd of the target (STDOUT_FILENO or STDERR_FILENO)
 * @return 0 on success, -1 on error
 */
static int receive_stream(int size,
			  int fd)
{
	if (size < 0 || size > BUFSIZE) {
		fprintf(stderr, "Invalid packet size\n");
		return -1;
	}

	if (read_buf(sd, &buf, size) < 0) {
		fprintf(stderr, "Can't read packet data from stream\n");
		return -1;
	}

	if (write(fd, &buf, size) != size) {
		perror("Can't write to stdout/stderr");
		return -1;
	}

	return 0;
}

/**
 * Reads a packet from the socket (sd) and writes it out if it's of type
 * STDOUT or STDERR. The value of an RC packet is stored in the global rc
 * variable.
 * @return 1 on RC, 0 on STDOUT/STDERR, -1 on error
 */
static int receive_packet(void)
{
	phead_t head;

	if (read_phead(sd, &head) < 0) {
		fprintf(stderr, "Can't read packet header from socket\n");
		return -1;
	}

	switch (head.type) {
	case PTYPE_STDOUT:
		return receive_stream(head.size, STDOUT_FILENO);

	case PTYPE_STDERR:
		return receive_stream(head.size, STDERR_FILENO);

	case PTYPE_RC:
		rc = read_int16(sd);
		if (rc < 0) {
			fprintf(stderr, "Can't read RC packet from socket\n");
			return -1;
		}
		return 1;

	case PTYPE_ERROR:
		/* Let's write the error message to stderr and add newline. */
		if (receive_stream(head.size, STDERR_FILENO) == 0) {
			fprintf(stderr, "\n");
		} else {
			fprintf(stderr, "Can't read message from server\n");
		}
		return -1;

	default:
		fprintf(stderr, "Received packet has invalid type\n");
		return -1;
	}
}

/**
 * Manages stdin/stdout/stderr and waits for the return code.
 * @return 0 on success, -1 on error
 */
static int manage(void)
{
	fd_set fds;

	while (1) {
		FD_ZERO(&fds);
		FD_SET(STDIN_FILENO, &fds);
		FD_SET(sd, &fds);

		if (select(sd + 1, &fds, NULL, NULL, NULL) <= 0) {
			perror("Can't do select() on stdin and socket");
			return -1;
		}

		if (FD_ISSET(STDIN_FILENO, &fds)) {
			if (send_packet() < 0) {
				return -1;
			}
		}

		if (FD_ISSET(sd, &fds)) {
			int ok;
			
			ok = receive_packet();
			if (ok < 0) {
				return -1;
			}
			if (ok > 0) {
				return 0;
			}
			/* else continue */
		}
	}
}

/**
 * Fills in the options from argv. Prints usage and exits on error.
 */
static void get_options(int argc,
			char **argv,
			config_t *cfg)
{
	if (argc < 6) {
		fprintf(stderr, "Usage: %s host user root pwd cmd [args]\n",
			argv[0]);
		exit(1);
	}

	cfg->host = argv[1];
	cfg->user = argv[2];
	cfg->root = argv[3];
	cfg->dir = argv[4];
	cfg->cmd = argv[5];
	cfg->argv = &argv[5];  /* Include the command name! */
}

/**
 * Fills in the port and password from ~/.sbrsh
 */
static void read_config(config_t *cfg)
{
	char *home, *path;

	home = getenv("HOME");
	path = malloc(strlen(home) + strlen("/" CONFIG_NAME) + 1);
	strcpy(path, home);
	strcat(path, "/" CONFIG_NAME);

	cfg->port = 0;
	cfg->pwd = find_pwd(path, cfg->host, &cfg->port);

	free(path);

	if (cfg->port <= 0 || cfg->port >= 65536) {
		fprintf(stderr, "Invalid port number: %d\n", cfg->port);
		exit(1);
	}
}

int main(int argc,
	 char **argv)
{
	config_t cfg;
	struct hostent *host;
	struct sockaddr_in addr;
	int16_t authok;

	get_options(argc, argv, &cfg);
	read_config(&cfg);

	/* Lookup host. */
	host = gethostbyname(cfg.host);
	if (!host) {
		fprintf(stderr, "Can't resolve host\n");
		return 1;
	}

	/* Create socket. */
	sd = socket(PF_INET, SOCK_STREAM, 0);
	if (sd < 0) {
		perror("Can't create socket");
		return 1;
	}

	/* Connect to server. */
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = *(unsigned int *) host->h_addr;
	addr.sin_port = htons(cfg.port);

	if (connect(sd, (struct sockaddr *) &addr, sizeof (addr)) < 0) {
		perror("Can't connect");
		return 1;
	}

	/* Send authentication info to server. */
	write_auth(sd, cfg.user, cfg.pwd);

	/* Wait for a reply... */
	authok = read_int16(sd);
	if (authok < 0) {
		fprintf(stderr, "Can't read authentication reply\n");
		return 1;
	}

	if (!authok) {
		fprintf(stderr, "Authentication failed :(\n");
		return 1;
	}

	/* Send command info to server. */
	write_cmd(sd, cfg.cmd, (const char **) cfg.argv, cfg.root, cfg.dir,
		  (const char **) environ);

	/* Manage stdin/out/err and wait for the rc. */
	if (manage() < 0) {
		return 1;
	}

	return rc;
}
 
