/*
 * $Id: daemon.c,v 1.3 2003/02/21 13:17:12 tsavola Exp $
 * Timo Savola <timo.savola@movial.fi>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <pwd.h>
#include <netdb.h>
#include <signal.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "daemon.h"
#include "protocol.h"
#include "auth.h"

#define DEBUG(msg) if (debugfile) debugfunc(msg)

/* The environment of the process. */
extern char **environ;

static FILE *debugfile = NULL;
static int sd;
static auth_info_t *auth;
static cmd_info_t *cmd;
static struct passwd *passwd;
static int inpipe[2];
static int outpipe[2];
static int errpipe[2];
static char buf[BUFSIZE];

/**
 * Print pid and message.
 * TODO: print time?
 */
static void debugfunc(char *msg)
{
	fprintf(debugfile, "[%5d] %s\n", getpid(), msg);
}

/**
 * Executes the command (in a child-child process).
 */
static void execute(void)
{
	/* We don't need this in here. */
	close(sd);

	/* Change stdin. */
	close(inpipe[1]);
	if (dup2(inpipe[0], STDIN_FILENO) != STDIN_FILENO) {
		syslog(LOG_ERR, "can't duplicate pipe as stdin");
		return;
	}

	/* Change stdout. */
	close(outpipe[0]);
	if (dup2(outpipe[1], STDOUT_FILENO) != STDOUT_FILENO) {
		syslog(LOG_ERR, "can't duplicate pipe as stdout");
		return;
	}

	/* Change stderr. */
	close(errpipe[0]);
	if (dup2(errpipe[1], STDERR_FILENO) != STDERR_FILENO) {
		syslog(LOG_ERR, "can't duplicate pipe as stderr");
		return;
	}

	/* Change environment. */
	environ = cmd->env;

	/* Just do it. */
	if (execvp(cmd->cmd, cmd->argv) < 0) {
		syslog(LOG_ERR, "can't execute command");
	}
}

/**
 * Reads some data from a pipe and writes it into the socket (sd) in a packet.
 * @param pipe the stdout or stderr of the child process
 * @param ptype the packet type
 * @return 0 if some data was copied, 1 if the pipe hit EOF, -1 on error
 */
static int send_packet(int pipe,
		       int ptype)
{
	int count;

	count = read(pipe, &buf, BUFSIZE);
	if (count < 0) {
		syslog(LOG_ERR, "can't read from pipe");
		return -1;
	}

	if (count > 0) {
		if (write_packet(sd, ptype, &buf, count) < 0) {
			syslog(LOG_ERR, "can't write packet to socket");
			return -1;
		}
		return 0;
	} else {
		return 1;
	}
}

/**
 * Reads a packet of type STDIN or STDIN_EOF from the socket (sd) and does
 * something rational about it.
 * @param pipe the stdin of the child process
 * @return 0 on success, -1 on error
 */
static int receive_packet(int pipe)
{
	phead_t head;

	if (read_phead(sd, &head) < 0) {
		syslog(LOG_ERR, "can't read packet header from socket");
		return -1;
	}

	switch (head.type) {
	case PTYPE_STDIN:
		if (read_buf(sd, &buf, head.size) < 0) {
			syslog(LOG_ERR, "can't read packet data from socket");
			return -1;
		}
		if (write(pipe, &buf, head.size) != head.size) {
			syslog(LOG_ERR, "can't write data to pipe");
			return -1;
		}
		return 0;

	case PTYPE_STDIN_EOF:
		close(pipe);
		return 0;

	default:
		syslog(LOG_ERR, "invalid packet type");
		return -1;
	}
}

/**
 * Manages the command executor (= the child-child process).
 * @param pid of the child process, used for getting the exit status
 */
static void manage(pid_t pid)
{
	int maxfd;
	fd_set fds;
	int outeof = 0;
	int erreof = 0;
	int status = 0;
	int rc = 0;
	uint16_t data;

	/* These are not for us. */
	close(inpipe[0]);
	close(outpipe[1]);
	close(errpipe[1]);

	/* Find out the maximum fd for select; this may be dumb but so am I. */
	maxfd = outpipe[0];
	if (errpipe[0] > maxfd) {
		maxfd = errpipe[0];
	}
	if (sd > maxfd) {
		maxfd = sd;
	}
	maxfd++;

	/* Copy in/out/err around while the child has out/err open. */
	while (!outeof || !erreof) {
		FD_ZERO(&fds);
		FD_SET(sd, &fds);
		if (!outeof) {
			FD_SET(outpipe[0], &fds);
		}
		if (!erreof) {
			FD_SET(errpipe[0], &fds);
		}
	
		if (select(maxfd, &fds, NULL, NULL, NULL) <= 0) {
			syslog(LOG_ERR, "can't select");
			return;
		}

		if (FD_ISSET(sd, &fds)) {
			DEBUG("receiving stdin packet");
			if (receive_packet(inpipe[1]) < 0) {
				return;
			}
		}

		if (FD_ISSET(outpipe[0], &fds)) {
			DEBUG("sending stdout packet");
			outeof = send_packet(outpipe[0], PTYPE_STDOUT);
		}

		if (FD_ISSET(errpipe[0], &fds)) {
			DEBUG("sending stderr packet");
			erreof = send_packet(errpipe[0], PTYPE_STDERR);
		}
	}

	/* Wait for the exit status of the executor process. */
	if (waitpid(pid, &status, 0) != pid) {
		syslog(LOG_ERR, "can't get exit status from the child");
		return;
	}

	rc = 1;
	if (WIFEXITED(status)) {
		DEBUG("command exited normally");
		rc = WEXITSTATUS(status);
	} else {
		DEBUG("command exited abnormally");
	}

	DEBUG("sending rc packet");

	/* Send the return code to the client. */
	data = htons(rc);
	write_packet(sd, PTYPE_RC, &data, sizeof (data));

	/* Let's wait until the client has gone away... */
	FD_ZERO(&fds);
	FD_SET(sd, &fds);
	select(sd + 1, NULL, NULL, &fds, NULL);
}

/**
 * @param host of the client
 * @return the truth
 */
static int authentic(char *host)
{
	char *path, *pwd;
	int ok;

	path = malloc(strlen(passwd->pw_dir) + strlen("/" CONFIG_NAME) + 1);
	strcpy(path, passwd->pw_dir);
	strcat(path, "/" CONFIG_NAME);

	pwd = find_pwd(path, host, NULL);
	if (!pwd) {
		syslog(LOG_WARNING, "can't find password from config file");
		return 0;
	}

	ok = (strcmp(pwd, auth->pwd) == 0);

	free(pwd);

	if (!ok) {
		syslog(LOG_WARNING, "the passwords don't match");
	}
	return ok;
}

/**
 * Handles one client connection (in a child process).
 * @param host of the client
 */
static void handle(char *host)
{
	pid_t pid;

	DEBUG("reading auth info struct from socket");

	/* Read the authentication info. */
	auth = read_auth(sd);
	if (!auth) {
		syslog(LOG_ERR, "can't read authentication info");
		return;
	}

	DEBUG("getting passwd struct of the user");

	/* Find out the user's uid and home. */
	passwd = getpwnam(auth->user);
	if (!passwd) {
		syslog(LOG_WARNING, "can't get user's passwd structure");

		/* Write negative authentication reply. */
		write_int16(sd, 0);
		return;
	}

	DEBUG("checking password");

	/* Check the password againts the config file. */
	if (!authentic(host)) {
		/* Write negative authentication reply. */
		write_int16(sd, 0);
		return;
	}

	DEBUG("writing positive auth reply to socket");

	/* Write positive authentication reply. */
	write_int16(sd, htons(1));

	DEBUG("reading command info struct from socket");

	/* Read the command info. */
	cmd = read_cmd(sd);
	if (!cmd) {
		syslog(LOG_ERR, "can't read command info");
		return;
	}

	DEBUG("verifying root directory");

	/* Check that root is under user's home. (This method is dubious.) */
	if (strstr(cmd->root, passwd->pw_dir) != cmd->root) {
		char *msg = "root directory is not under home";

		syslog(LOG_WARNING, msg);

		/* Tell it to the client aswell. */
		write_packet(sd, PTYPE_ERROR, msg, strlen(msg));

		return;
	}

	DEBUG("changing root directory");

	/* Change root directory. We need to be root to do this. */
	if (chroot(cmd->root) < 0) {
#ifdef NO_CHROOT
		syslog(LOG_WARNING, "can't change root directory");
#else
		syslog(LOG_ERR, "can't change root directory");
		return;
#endif
	}

	DEBUG("changing current directory");

	/* Change pwd. */
	if (chdir(cmd->dir) < 0) {
		syslog(LOG_ERR, "can't change current directory");
		return;
	}

	DEBUG("changing uid");

	/* Change uid to that of the user. */
	if (setuid(passwd->pw_uid) < 0) {
		syslog(LOG_ERR, "can't change uid");
		return;
	}

	DEBUG("creating pipes");

	/* Manager copies inpipe from and outpipe/errpipe to the client. */
	if (pipe(inpipe) < 0 ||
	    pipe(outpipe) < 0 ||
	    pipe(errpipe) < 0) {
		syslog(LOG_ERR, "can't create pipes");
		return;
	}

	DEBUG("creating the command process");

	pid = fork();
	if (pid < 0) {
		syslog(LOG_ERR, "can't create command process");
		return;
	}

	if (pid == 0) {
		/* Execute the command in the child process. */
		execute();
	} else {
		/* Manage the stdin/stdout/stderr of the child process. */
		manage(pid);
	}
}

/**
 * Parse args, open logs, create server socket, accept connections.
 */
int main(int argc,
	 char **argv)
{
	int port;
	int srvsd;
	struct sockaddr_in srvaddr;

	if (argc < 2 || argc > 3) {
		fprintf(stderr, "Usage: %s port [debugfile]\n", argv[0]);
		return 1;
	}

	openlog(argv[0], LOG_PERROR | LOG_PID, LOG_DAEMON);

	port = atoi(argv[1]);
	if (port <= 0 || port >= 65536) {
		syslog(LOG_ERR, "invalid port number: %d", port);
		return 1;
	}

	/* Open debug log? */
	if (argc > 2) {
		if (strcmp(argv[2], "-") == 0) {
			debugfile = stdout;
		} else {
			debugfile = fopen(argv[2], "a");
			if (!debugfile) {
				syslog(LOG_WARNING, "can't open debug log");
			}
		}
	}

	/* Create a socket, bind it and listen to it. */
	srvsd = socket(PF_INET, SOCK_STREAM, 0);
	if (srvsd < 0) {
		syslog(LOG_ERR, "can't create socket");
		return 1;
	}

	srvaddr.sin_family = AF_INET;
	srvaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	srvaddr.sin_port = htons(port);

	if (bind(srvsd, (struct sockaddr *) &srvaddr, sizeof (srvaddr)) < 0) {
		syslog(LOG_ERR, "can't bind to port: %d", port);
		return 1;
	}

	if (listen(srvsd, BACKLOG) < 0) {
		syslog(LOG_ERR, "can't listen");
		return 1;
	}

	/* Accept connections from clients. */
	while (1) {
		struct sockaddr_in addr;
		socklen_t len;
		char host[NI_MAXHOST];
		pid_t pid;

		DEBUG("waiting for connection");

		sd = accept(srvsd, (struct sockaddr *) &addr, &len);
		if (sd < 0) {
			syslog(LOG_ERR, "can't accept connection");
			continue;
		}

		DEBUG("new connection received");

		/* Get the IP address. */
		if (getnameinfo((struct sockaddr *) &addr, len,
				host, sizeof(host),
				NULL, 0, NI_NUMERICHOST) != 0) {
			syslog(LOG_ERR, "can't get client's ip address");
			continue;
		}

		DEBUG("creating handler process");

		pid = fork();

		/* Handle everything in the child. */
		if (pid == 0) {
			close(srvsd);
			handle(host);
			return 0;
		}

		if (pid < 0) {
			syslog(LOG_ERR, "can't create handler process");
		}

		close(sd);
	}

	return 0;
}
